from datetime import date, datetime, time
from typing import Optional, Type

from _decimal import Decimal

from qtgql.codegen.py.scalars import BaseCustomScalar


class DateTimeScalar(BaseCustomScalar[Optional[datetime]]):
    """An ISO-8601 encoded datetime."""

    GRAPHQL_NAME: str = "DateTime"
    DEFAULT_DESERIALIZED = " --- "

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "DateTimeScalar":
        if v:
            return cls(datetime.fromisoformat(v))
        return cls(None)

    def to_qt(self) -> str:
        if self._value:
            return self._value.strftime(" %H:%M: ")
        return self.DEFAULT_DESERIALIZED


class DateScalar(BaseCustomScalar[Optional[date]]):
    """An ISO-8601 encoded date."""

    GRAPHQL_NAME = "Date"

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "DateScalar":
        if v:
            return cls(date.fromisoformat(v))
        return cls(None)

    def to_qt(self) -> str:
        if self._value:
            return self._value.isoformat()
        return self.DEFAULT_DESERIALIZED


class TimeScalar(BaseCustomScalar[Optional[time]]):
    """an ISO-8601 encoded time."""

    GRAPHQL_NAME = "Time"

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "TimeScalar":
        if v:
            return cls(time.fromisoformat(v))
        return cls(None)

    def to_qt(self) -> str:
        if self._value:
            return self._value.isoformat()
        return self.DEFAULT_DESERIALIZED


class DecimalScalar(BaseCustomScalar[Decimal]):
    """A Decimal value serialized as a string."""

    GRAPHQL_NAME = "Decimal"
    DEFAULT_DESERIALIZED = Decimal()

    @classmethod
    def from_graphql(cls, v: Optional[str] = None) -> "DecimalScalar":
        if v:
            return cls(Decimal(v))
        return cls(cls.DEFAULT_DESERIALIZED)

    def to_qt(self) -> str:
        return str(self._value)


CustomScalarMap = dict[str, Type[BaseCustomScalar]]
CUSTOM_SCALARS: CustomScalarMap = {
    DateTimeScalar.GRAPHQL_NAME: DateTimeScalar,
    DecimalScalar.GRAPHQL_NAME: DecimalScalar,
    DateScalar.GRAPHQL_NAME: DateScalar,
    TimeScalar.GRAPHQL_NAME: TimeScalar,
}



from __future__ import annotations

from typing import Optional, TypeVar

from PySide6.QtCore import QAbstractListModel, QByteArray, QObject, Qt

from qtgql import slot

__all__ = ["BaseModel", "get_base_graphql_object"]


class _BaseQGraphQLObject(QObject):
    type_map: dict[str, type[_BaseQGraphQLObject]]

    def __init_subclass__(cls, **kwargs):
        cls.type_map[cls.__name__] = cls

    def __init__(self, parent: Optional[QObject] = None):
        super().__init__(parent)

    @classmethod
    def from_dict(
        cls, parent: T_BaseQGraphQLObject, data: dict
    ) -> T_BaseQGraphQLObject:  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def deserialize_optional_child(
        cls,
        parent: Optional[T_BaseQGraphQLObject],
        data: dict,
        child: type[T_BaseQGraphQLObject],
        field_name: str,
    ) -> Optional[T_BaseQGraphQLObject]:
        if found := data.get(field_name, None):
            return child.from_dict(parent, found)  # type: ignore

    @classmethod
    def deserialize_list_of(
        cls,
        parent: Optional[T_BaseQGraphQLObject],
        data: dict,
        model: type[T_BaseModel],
        field_name: str,
        of_type: type[T_BaseQGraphQLObject],
    ) -> Optional[T_BaseModel]:
        if found := data.get(field_name, None):
            return model(parent=parent, data=[of_type.from_dict(parent, data) for data in found])  # type: ignore

    @classmethod
    def deserialize_union(
        cls,
        parent: Optional[T_BaseQGraphQLObject],
        data: dict,
        field_name: str,
    ) -> Optional[T_BaseModel]:
        if found := data.get(field_name, None):
            child = cls.type_map[found["__typename"]]
            return child.from_dict(parent, found)  # type: ignore


class BaseModel(QAbstractListModel):
    OBJECT_ROLE = Qt.ItemDataRole.UserRole + 1

    def __init__(self, data: list[T_BaseQGraphQLObject], parent: Optional[QObject] = None):
        super().__init__(parent)
        self._data = data

    def rowCount(self, *args, **kwargs) -> int:
        return len(self._data)

    def roleNames(self) -> dict:
        return {self.OBJECT_ROLE: QByteArray("object")}  # type: ignore

    def data(self, index, role=...) -> Optional[T_BaseQGraphQLObject]:
        if index.row() < len(self._data) and index.isValid():
            if role == self.OBJECT_ROLE:
                return self._data[index.row()]
            raise NotImplementedError(
                f"role {role} is not a valid role for {self.__class__.__name__}"
            )

    def append(self, node: T_BaseQGraphQLObject) -> None:
        count = self.rowCount()
        self.beginInsertRows(self.index(count), count, count)
        self._data.append(node)
        self.endInsertRows()

    @slot
    def pop(self, index: Optional[int] = None) -> None:
        index = -1 if index is None else index
        real_index = index if index > -1 else self.rowCount()
        self.beginRemoveRows(self.index(index - 1).parent(), real_index, real_index)
        self._data.pop(index)
        self.endRemoveRows()


def get_base_graphql_object(name: str) -> type[_BaseQGraphQLObject]:
    """
    :param name: valid attribute name (used by codegen to import it).
    :returns: A type to be extended by all generated types.
    """
    type_map: dict[str, type[_BaseQGraphQLObject]] = {}
    return type(name, (_BaseQGraphQLObject,), {"type_map": type_map})  # type: ignore


BaseGraphQLObject = get_base_graphql_object("BaseGraphQLObject")

T_BaseModel = TypeVar("T_BaseModel", bound=BaseModel)
T_BaseQGraphQLObject = TypeVar("T_BaseQGraphQLObject", bound=_BaseQGraphQLObject)



from __future__ import annotations

import contextlib
import enum
from functools import cached_property
from typing import List, Optional, Type, Union

from attrs import define

from qtgql.codegen.py.bases import _BaseQGraphQLObject
from qtgql.codegen.py.custom_scalars import CustomScalarMap
from qtgql.codegen.py.scalars import BaseCustomScalar, BuiltinScalars
from qtgql.codegen.utils import AntiForwardRef
from qtgql.typingref import TypeHinter, ensure


class Kinds(enum.Enum):
    SCALAR = "SCALAR"
    OBJECT = "OBJECT"
    ENUM = "ENUM"
    LIST = "LIST"
    UNION = "UNION"
    NON_NULL = "NON_NULL"
    INTERFACE = "INTERFACE"
    INPUT_OBJECT = "INPUT_OBJECT"


@define(slots=False)
class FieldProperty:
    name: str
    type: TypeHinter
    type_map: dict[str, GqlTypeDefinition]
    enums: "EnumMap"
    scalars: CustomScalarMap
    description: Optional[str] = ""

    @cached_property
    def deserializer(self) -> str:
        """If the field is optional this would be."""
        # every thing is possibly optional since you can query for only so or so fields.
        default = f"data.get('{self.name}', None)"
        t = self.type.type
        type_hinter = self.type
        if t is Optional:
            t = self.type.of_type[0].type
            type_hinter = self.type.of_type[0]

        if t in BuiltinScalars.values():
            return default

        if scalar := self.is_custom_scalar:
            return f"SCALARS.{scalar.__name__}.{BaseCustomScalar.from_graphql.__name__}({default})"
        if t in (list, List):
            inner = type_hinter.of_type[0].type
            # handle inner type
            assert issubclass(inner, AntiForwardRef)
            # enums are not supported in lists yet (valid graphql spec though)
            gql_type = ensure(inner.resolve(), GqlTypeDefinition)
            assert gql_type
            return f"cls.{_BaseQGraphQLObject.deserialize_list_of.__name__}(parent, data, {gql_type.model_name}, '{self.name}', {gql_type.name})"
        if t is Union:
            return (
                f"cls.{_BaseQGraphQLObject.deserialize_union.__name__}(parent, data, '{self.name}')"
            )
        # handle inner type if it was optional above
        assert issubclass(t, AntiForwardRef)
        # graphql object
        if gql_type := self.is_object_type:
            return f"cls.{_BaseQGraphQLObject.deserialize_optional_child.__name__}(parent, data, {gql_type.name}, '{self.name}')"
        elif isinstance(t.resolve(), GqlEnumDefinition):
            return f"{t.name}[data.get('{self.name}', 1)]"  # graphql enums evaluates to string of the name.
        raise NotImplementedError  # pragma: no cover

    @cached_property
    def annotation(self) -> str:
        """
        :returns: Annotation of the field based on the real type,
        meaning that the private attribute would be of that type.
        this goes for init and the property setter.
        """
        ret = self.type.as_annotation()
        # int, str, float etc...
        if ret in BuiltinScalars.values():
            return ret.__name__

        if scalar := self.is_custom_scalar:
            return f"SCALARS.{scalar.__name__}"
        if gql_enum := self.is_enum:
            return gql_enum.name
        # handle Optional, Union, List etc...
        # removing redundant prefixes.
        return TypeHinter.from_annotations(ret).stringify()

    @cached_property
    def fget_annotation(self) -> str:
        """This annotates the value that is QML-compatible."""
        ret = self.type
        if ret.is_optional():
            ret = self.type.of_type[0]

        if ret.type in BuiltinScalars.values():
            return self.annotation
        if scalar := self.is_custom_scalar:
            return TypeHinter.from_annotations(scalar.to_qt.__annotations__["return"]).stringify()
        if ret.is_list():
            return self.type_map[ret.of_type[0].type.name].model_name
        if self.is_enum:
            return "int"

        return ret.stringify()

    @cached_property
    def property_type(self) -> str:
        try:
            # this should raise if it is an inner type.
            ret = TypeHinter.from_string(self.fget_annotation, self.type_map)
            if ret.is_optional():
                return ret.of_type[0].stringify()
            if ret.type in self.type_map.values():
                raise TypeError  # in py3.11 the get_type_hints won't raise so raise brutally
            return self.fget_annotation
        except (TypeError, NameError):
            if self.type.is_union():
                return "QObject"  # graphql doesn't support scalars in Unions ATM.
            if self.is_enum:
                # QEnum value must be int
                return "int"
            # might be a model, which is also QObject
            assert self.is_model or self.is_object_type
            return "QObject"

    @cached_property
    def setter_name(self) -> str:
        return self.name + "_setter"

    @cached_property
    def signal_name(self) -> str:
        return self.name + "Changed"

    @cached_property
    def private_name(self) -> str:
        return "_" + self.name

    @staticmethod
    def unwrap_optional(th: TypeHinter) -> TypeHinter:
        if th.is_optional():
            return th.of_type[0]
        return th

    @cached_property
    def is_object_type(self) -> Optional[GqlTypeDefinition]:
        t = self.type.type
        if self.type.is_optional():
            t = self.type.of_type[0].type
        with contextlib.suppress(TypeError):
            if issubclass(t, AntiForwardRef):
                ret = t.resolve()
                if isinstance(ret, GqlTypeDefinition):
                    return ret

    @cached_property
    def is_model(self) -> Optional[GqlTypeDefinition]:
        th = self.type
        if th.is_optional():
            th = th.of_type[0]
        if th.is_list():
            th = th.of_type[0]
        with contextlib.suppress(TypeError):
            if issubclass(th.type, AntiForwardRef):
                ret = th.type.resolve()
                if isinstance(ret, GqlTypeDefinition):
                    return ret

    @cached_property
    def is_scalar(self) -> bool:
        return (
            self.type.type in BuiltinScalars.values()
            or self.type.of_type in BuiltinScalars.values()
            and not self.is_custom_scalar
        )

    @cached_property
    def is_enum(self) -> Optional["GqlEnumDefinition"]:
        tp = self.unwrap_optional(self.type)
        with contextlib.suppress(TypeError):
            if issubclass(tp.type, AntiForwardRef):
                if isinstance(tp.type.resolve(), GqlEnumDefinition):
                    return tp.type.resolve()

    @cached_property
    def is_custom_scalar(self) -> Optional[Type[BaseCustomScalar]]:
        tp = self.unwrap_optional(self.type)
        if tp.type in self.scalars.values():
            return tp.type

    @property
    def fget(self) -> str:
        if self.is_custom_scalar:
            return f"return self.{self.private_name}.{BaseCustomScalar.to_qt.__name__}()"
        if self.is_enum:
            return f"return self.{self.private_name}.value"
        return f"return self.{self.private_name}"


@define(slots=False)
class GqlTypeDefinition:
    kind: Kinds
    name: str
    fields: list[FieldProperty]
    docstring: Optional[str] = ""

    @cached_property
    def model_name(self) -> str:
        return "Q" + self.name + "Model"


@define
class EnumValue:
    """encapsulates enumValues from introspection, maps to an Enum member."""

    name: str
    description: str = ""


@define
class GqlEnumDefinition:
    name: str
    members: list[EnumValue]


EnumMap = dict[str, "GqlEnumDefinition"]



from typing import Any, Generic, Optional, TypeVar

T = TypeVar("T")

__all__ = ["BaseCustomScalar", "BuiltinScalars"]


class BaseCustomScalar(Generic[T]):
    """Class to extend by user defined scalars."""

    __slots__ = "_value"
    GRAPHQL_NAME: str
    """The *real* GraphQL name of the scalar (used by the codegen inspection
    pipeline)."""
    DEFAULT_DESERIALIZED: Any = " --- "
    """A place holder graphql query returned null or the field wasn't queried.

    can be used by `from_graphql()`
    """

    def __init__(self, v: T):
        self._value = v

    @classmethod
    def from_graphql(cls, v: Optional[Any] = None) -> "BaseCustomScalar":
        """Deserializes data fetched from graphql, This is useful when you want
        to set a first-of value that will later be used by `to_qt()`.

        **must be overridden**!
        """
        raise NotImplementedError  # pragma: no cover

    def to_qt(self) -> Any:
        """Will be used by the property getter, This is the official value that
        Qt should "understand".

        **must be overridden**!
        """
        raise NotImplementedError  # pragma: no cover


BuiltinScalars: dict[str, type] = {
    "Int": int,
    "Float": float,
    "String": str,
    "ID": str,
    "Boolean": bool,
    "UUID": str,
}



import itertools
from typing import Any, Optional, get_args

from PySide6 import QtCore as qtc

from qtgql.typingref import is_optional


def get_optional_args(annotations: list[Any]) -> list[Any]:
    return [get_args(arg)[0] for arg in annotations if is_optional(arg)]


def get_concrete(annotation: Any) -> Optional[type]:
    return getattr(annotation, "__origin__", None)


def get_concretes(annotations: list) -> list:
    ret = []
    for item in annotations:
        if is_optional(item):
            annotations.remove(item)
            continue
        if concrete := get_concrete(item):
            ret.append(concrete)
            annotations.remove(item)
    return ret + annotations


def slot(func):
    def wrapper(func) -> func:
        anots: dict = func.__annotations__
        return_ = anots.pop("return", None)
        if return_ is Any:
            return_ = "QVariant"
        args = list(anots.values())
        stripped_optionals = get_concretes(get_optional_args(args))
        concretes = get_concretes(args)

        if stripped_optionals:
            required_annotations = tuple(concretes)
            combos = []
            for i in range(len(stripped_optionals) + 1):
                for subset in itertools.combinations(stripped_optionals, i):
                    combos.append(subset)
            for combo in combos:
                func = qtc.Slot(*combo + required_annotations, result=return_)(func)

            return func

        return qtc.Slot(*concretes, result=return_)(func)

    return wrapper(func)



import inspect
from typing import Any, List, Optional, Type, TypeVar, Union, get_args, get_origin

from attrs import define


def is_optional(field) -> bool:
    return get_origin(field) is Union and type(None) in get_args(field)


class UnsetType:
    __instance: Optional["UnsetType"] = None

    def __new__(cls: Type["UnsetType"]) -> "UnsetType":
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
            return cls.__instance
        return cls.__instance

    def __str__(self):
        return ""

    def __repr__(self) -> str:  # pragma: no cover
        return "UNSET"

    def __bool__(self):
        return False


UNSET: Any = UnsetType()


@define
class TypeHinter:
    type: Any  # noqa: A003
    of_type: tuple["TypeHinter", ...] = ()

    @classmethod
    def from_string(cls, tp: str, ns: dict) -> "TypeHinter":
        from typing import List, Union, Dict, Optional, Any, Type, get_type_hints  # noqa

        ns = ns.copy()
        ns.update(locals())

        def tmp(t: tp):  # type: ignore
            ...

        annotation = get_type_hints(tmp, localns=ns, globalns=globals())["t"]

        return cls.from_annotations(annotation)

    @classmethod
    def from_annotations(cls, tp: Any) -> "TypeHinter":
        if args := get_args(tp):
            # handle optional
            if type(None) in args:
                return TypeHinter(type=Optional, of_type=(TypeHinter.from_annotations(args[0]),))
            new_args: list[TypeHinter] = []
            for arg in args:
                new_args.append(TypeHinter.from_annotations(arg))

            return TypeHinter(type=get_origin(tp), of_type=tuple(new_args))  # type: ignore
        return TypeHinter(type=tp)

    def as_annotation(self, object_map: Optional[dict[str, Any]] = None) -> Any:
        if self.type is str:
            return self.type
        # eval forward refs
        if isinstance(self.type, str):
            assert object_map, "can't evaluate forward refs without object_map."
            self.type = object_map[self.type]

        if builder := getattr(
            self.type, "__class_getitem__", getattr(self.type, "__getitem__", None)
        ):
            if self.is_generic():
                return self.type

            if self.is_union():
                return builder(tuple(arg.as_annotation(object_map) for arg in self.of_type))
            return builder(self.of_type[0].as_annotation(object_map))
        return self.type

    def stringify(self) -> str:
        annot = self.as_annotation()
        if inspect.isclass(annot) and not hasattr(annot, "__origin__"):
            return annot.__name__

        return str(annot).replace("typing.", "").replace("qtgql.codegen.utils.", "")

    def is_union(self) -> bool:
        return self.type is Union

    def is_optional(self) -> bool:
        return self.type is Optional

    def is_generic(self) -> bool:
        return inspect.isclass(self.type) and "Generic.__class_getitem__" in repr(
            self.type.__class_getitem__
        )

    def is_list(self) -> bool:
        return self.type in (list, List)


T = TypeVar("T")


def ensure(inst: T, tp: Type[T]) -> T:
    if not isinstance(inst, tp):
        raise TypeError(f"{inst} is not of type {tp}")
    return inst
